import os
import sys
import re
from pathlib import Path
import yaml
from jsonschema import validate
import importlib
from rdflib import Namespace
from rdflib import Graph
from rdflib.namespace import NamespaceManager
from rdflib.compare import to_isomorphic, graph_diff


BASE = "https://saref.etsi.org/"

class SAREFProject:
    def __init__(self, projectRepository : Path):
        repositoryName = os.path.basename(projectRepository)
        if repositoryName == "saref-core":
            self.projectRepository = projectRepository
            self.repositoryName = repositoryName
            self.name = "SAREF"
            self.prefix = "saref"
            self.path = "core"
            self.namespace = Namespace(f"{BASE}{self.path}/")
            self.ontologyFileName = "saref"
        elif re.match(r"^saref4[a-z]{4}$", repositoryName):
            self.projectRepository = projectRepository
            self.repositoryName = repositoryName
            self.acronym = repositoryName[6:] # type: str
            self.name = f"SAREF4{self.acronym.upper()}"
            self.prefix = f"s4{self.acronym}"
            self.path = f"saref4{self.acronym}"
            self.namespace = Namespace(f"{BASE}{self.path}/")
            self.ontologyFileName = f"saref4{self.acronym}"
        else:
            raise ValueError("The name of the repository does not conform to SAREF Project names")

SAREFCore = SAREFProject("saref-core")
SAREF = SAREFCore.namespace

class SAREFPattern:
    def __init__(self, specialize, schema):
        self.specialize = specialize     
        self.schema = schema

def run_one(project, pattern, executions):
    g = Graph()
    g += pattern.ontology
    for execution in executions:
        # check the format 
        if pattern.schema:
            validate(execution, pattern.schema)
        # execute the pattern specialization function
        g += pattern.specialize(project.namespace, execution)

    # check if some triples are not in the ontology 
    g = to_isomorphic(g) 
    in_both, in_first, in_second = graph_diff(project.ontology, g)
    if in_second:
        sys.stderr.write("Error: the following triples should be in the ontology:\n")
        in_second.namespace_manager = project.namespace_manager
        in_second.print(out=sys.stderr)

def get_project(patterns_config_yml:Path):
    projectRepository = patterns_config_yml.parents[1]
    project = SAREFProject(projectRepository)
    
    project.namespace_manager = NamespaceManager(Graph())
    project.namespace_manager.bind(SAREFCore.prefix, SAREF)
    project.namespace_manager.bind(project.prefix, project.namespace)

    with open(projectRepository / "ontology" / (project.ontologyFileName + ".ttl"), 'r') as ontology_file:
        project.ontology = Graph()
        project.ontology.parse(ontology_file)
        project.ontology = to_isomorphic(project.ontology) 
    
    project.namespace_manager = NamespaceManager(Graph())
    project.namespace_manager.bind(SAREFCore.prefix, SAREF)
    project.namespace_manager.bind(project.prefix, project.namespace)
    return project

def check_pattern(specialize, schema=None):
    if len(sys.argv) != 2:
        print()
        sys.stderr.write("Expecting one argument: the location of a pattern specialization specification document. See ETSI TS 103 673")
        sys.exit(1)
    scriptname = Path(sys.argv[0]).resolve()
    reference_pattern = os.path.basename(os.path.dirname(scriptname))
    project_with_reference_patterns = os.path.basename(scriptname.parents[2])
    pattern = SAREFPattern(specialize, schema)
    pattern.ontology = Graph()
    try:
        with open((scriptname/".."/"pattern.ttl").resolve(), "r") as pattern_ontology_file:
            pattern.ontology.parse(pattern_ontology_file)
    except Exception as e:
        print(e)
        pass
    
    patterns_config_yml = Path(sys.argv[1]).resolve()
    project = get_project(patterns_config_yml)
    
    with open(patterns_config_yml, 'r') as patterns_config_file:
        patterns_config = yaml.safe_load(patterns_config_file)
        
        if project_with_reference_patterns not in patterns_config:
            sys.stderr.write(f"Warning: Patterns specialization specification document does not contain a section for {project_with_reference_patterns}\n")
            sys.exit()
        if reference_pattern not in patterns_config[project_with_reference_patterns]:
            sys.stderr.write(f"Warning: Patterns specialization specification document does not contain a section for {project_with_reference_patterns}.{reference_pattern}\n")
            sys.exit()
        executions = patterns_config[project_with_reference_patterns][reference_pattern]
        run_one(project, pattern, executions)
    
if __name__ == "__main__":
    if len(sys.argv) != 2:
        print()
        sys.stderr.write("Expecting one argument: the location of a pattern specialization specification document. See ETSI TS 103 673")
        sys.exit(1)

    patterns_config_yml = Path(sys.argv[1]).resolve()
    project = get_project(patterns_config_yml)

    with open(patterns_config_yml, 'r') as patterns_config_file:
        patterns_config = yaml.safe_load(patterns_config_file)
        for project_with_reference_patterns in patterns_config:
            for reference_pattern in patterns_config[project_with_reference_patterns]:
                print(f"executing pattern {reference_pattern} in project {project_with_reference_patterns}")
                
                pattern_name = f"{project_with_reference_patterns}.{reference_pattern}"
                pattern_path = Path("..")/project_with_reference_patterns/"patterns"/reference_pattern
                spec = importlib.util.spec_from_file_location(pattern_name, pattern_path/"specialization.py")
                pattern = importlib.util.module_from_spec(spec)
                # sys.modules[module_name] = module
                spec.loader.exec_module(pattern)
                # initialize pattern.ontology
                pattern.ontology = Graph()
                try:
                    with open(pattern_path/"pattern.ttl", "r") as pattern_ontology_file:
                        pattern.ontology.parse(pattern_ontology_file)
                except:
                    pass
                executions = patterns_config[project_with_reference_patterns][reference_pattern]
                run_one(project, pattern, executions)
